from enum import Enum
from nmigen import *
from nmigen.hdl.rec import *
from nmigen.asserts import *
from nmigen.cli import main


# WARNING!
# Signal(range(8)) will generate a THREE(3) wide signal, not enough to store 8
# we need to check all our widths to verify we haven't fallen prey to this

# we need to harmonize the terminology of "left" and "right" with LSB and MSB type terminology


class ArbitraryWidthMemoryLayout(Layout):
    def __init__(self, *, data_width, address_width): # the * forces keyword args
        super().__init__([
            # DATA
            ("r_data",        unsigned(data_width)),      # OUTPUT

            # ADDRESS
            ("r_addr",        unsigned(address_width)),   # INPUT

            # FLOW CONTROL UPSTREAM
            ("valid_in",       1),                        # INPUT
            ("ready_out",      1),                        # OUTPUT

            # FLOW CONTROL DOWNSTREAM
            ("valid_out",      1),                        # OUTPUT
            ("ready_in",       1),                        # INPUT

        ])

class ArbitraryWidthMemoryBus(Record):
    def __init__(self, *, data_width, address_width):
        super().__init__(ArbitraryWidthMemoryLayout(data_width=data_width, address_width=address_width))


# initial_data must be folded, spindleed, and mutilated to fit inside the backing memory dimensions

# later on we will write a function to take care of this automatically


class ArbitraryWidthMemory(Elaboratable):
        def __init__(self, *, fake_data_width, fake_address_width, backing_memory_data_width, backing_memory_address_width, initial_data):
            self.fake_data_width    = fake_data_width
            self.fake_address_width = fake_address_width

            def is_power_of_two(n):
                return (n & (n-1) == 0)

            assert(is_power_of_two(backing_memory_data_width))

            self.backing_memory_data_width      = backing_memory_data_width
            self.backing_memory_data_width_bits = backing_memory_data_width.bit_length() - 1
            assert(2**self.backing_memory_data_width_bits == backing_memory_data_width)

            self.backing_memory_address_width = backing_memory_address_width

            self.backing_memory_length = 256 #len(initial_data)

            assert(2**backing_memory_address_width >= self.backing_memory_length)

            self.bus = ArbitraryWidthMemoryBus(data_width=fake_data_width, address_width=fake_address_width)
            self.backing_memory = Memory(width=backing_memory_data_width, depth=self.backing_memory_length, init=initial_data)

        def elaborate(self, platform):
            m = Module()

            bus = self.bus

            # regs to persist output of last read onto the output bus until it's accepted by the downstream consumer
            last_r_data       = Signal(self.fake_data_width)
            last_r_data_valid = Signal(1)



            unwrapped_bit_index = Signal(range(self.backing_memory_length*self.backing_memory_data_width)) #probably cannot be made shorter.

            left_bit_index = Signal(self.backing_memory_data_width)
            right_bit_index = Signal(self.backing_memory_data_width)
            end_bit_pseudo_index = Signal(range(self.backing_memory_length*self.backing_memory_data_width)) # can be made shorter
            additional_words = Signal(self.backing_memory_address_width) # can also be amde shorter

            starting_word = Signal(self.backing_memory_address_width)



            # State across clock cycles
            next_address = Signal(self.backing_memory_address_width)
            additional_words_regd = Signal(self.backing_memory_address_width)
            end_bit_pseudo_index_regd = Signal(range(self.backing_memory_length*self.backing_memory_data_width)) # can be made shorter

            # pseudo-output signals, we connect these to memory and a shift register later:
            valid_fetch = Signal(1)
            fetch_address = Signal(self.backing_memory_address_width)


            m.submodules.read_port  = read_port  = self.backing_memory.read_port()
            m.d.comb += read_port.addr.eq(fetch_address)
            m.d.comb += self.bus.r_data.eq(read_port.data)




            # combining shift reg

            # the register itself

            shreg = Signal(self.fake_data_width)
            current_slice = Signal(self.backing_memory_data_width)
            lower_bits_cut = Signal(self.backing_memory_data_width)

            valid_shreg_bits = Signal(range(self.fake_data_width))
            shreg_new_bits = Signal(range(self.backing_memory_data_width+1))

            m.d.sync += shreg.eq(shreg << (shreg_new_bits) | current_slice)

            m.d.comb += lower_bits_cut.eq(read_port.data>>left_bit_index)

            m.d.comb += current_slice.eq(((lower_bits_cut<<right_bit_index)&0xff)>>right_bit_index)

            m.d.comb += shreg_new_bits.eq(right_bit_index-left_bit_index+1)

#                is like a gear box but with variable upstream elnght

#                thats it

#                we should do this)





            with m.FSM() as fsm:
                with m.State("RESET"):
                    m.next ="READY"

                with m.State("READY"):
                    m.d.comb += bus.ready_out.eq(1)

                    with m.If(last_r_data_valid == 1):
                        m.d.comb += bus.valid_out.eq(1)
                        m.d.comb += bus.r_data.eq(last_r_data)

                    with m.If(bus.valid_in == 1):
                        # the memory address and bit-index computation goes as follows:

                        # We calculate an unwrapped bit index by multiplying the fake index by the fake data width
                        m.d.comb += unwrapped_bit_index.eq(bus.r_addr * self.fake_data_width)

                        # The memory address is simply the unwrapped bit index divided by the memory data width
                        # since the memory data width is a power of two, this is a shift!
                        m.d.comb += starting_word.eq(unwrapped_bit_index[self.backing_memory_data_width_bits:])

                        # We say we'll do a fetch at that address:
                        m.d.comb += fetch_address.eq(starting_word)

                        # We start our cut at the unwrapped bit index modulo the memory data width.
                        # since the memory data width is a power of two, this is the K-least-significant-bits
                        # of the unwrapped bit index
                        m.d.comb += left_bit_index.eq(unwrapped_bit_index[:self.backing_memory_data_width_bits])

                        # Here's where they start trying to trick you. We need to handle the case where the end of the
                        # fake word goes beyond a real memory word.
                        m.d.comb += end_bit_pseudo_index.eq(left_bit_index + self.fake_data_width - 1)

                        # So here we determine if there's any need for additional memory words:
                        m.d.comb += additional_words.eq(end_bit_pseudo_index[self.backing_memory_data_width_bits:])

                        with m.If(additional_words == 0):
                            # No additional words, calculate which bits we need from the sole word we're fetching:
                            m.d.comb += right_bit_index.eq(end_bit_pseudo_index[:self.backing_memory_data_width_bits])
                        with m.Else():
                            # Additional words needed, so we fetch the entire remaining part of this word
                            m.d.comb += right_bit_index.eq(self.backing_memory_data_width-1)
                            # and we register state for the next cycle[s]
                            m.d.sync += next_address.eq(fetch_address + 1)
                            # rather than keeping track of the next address and the final address,
                            # we keep track of the current address and how many additional words are left,
                            # maybe that saves a few LUTs by avoiding wide compares, who knows.
                            m.d.sync += additional_words_regd.eq(additional_words)
                            m.d.sync += end_bit_pseudo_index_regd.eq(end_bit_pseudo_index)
                            m.next="ADD"


                with m.State("ADD"):
                    m.d.comb += bus.ready_out.eq(0)

                    # we handle both the full-word fetches and the final (potentially partial word) fetch here
                    with m.If(additional_words_regd == 1):
                        # We start from zero...
                        m.d.comb += left_bit_index.eq(0)
                        # But this is the last word, so we may not have to include the whole word!
                        m.d.comb += right_bit_index.eq(end_bit_pseudo_index_regd[:self.backing_memory_data_width_bits])
                        m.d.comb += fetch_address.eq(next_address)
                        m.next = "STALL"
                    with m.Else():
                        # non-special case, fetch the whole word
                        m.d.comb += left_bit_index.eq(0)
                        m.d.comb += right_bit_index.eq(self.backing_memory_data_width-1)
                        # and increment the address and decrement the remaining words counter
                        m.d.sync += next_address.eq(next_address + 1)
                        m.d.sync += additional_words_regd.eq(additional_words_regd - 1)

                        m.d.comb += fetch_address.eq(next_address)
                with m.State("STALL"):
                    m.next="STALL"
            return m



# This is non-synthesizable but is intended to provide a model for formal verification.
# We initialize the "sim_memory" with the same data that's in the 


class GoldenArbitraryWidthMemory(Elaboratable):
        def __init__(self, *, data_width, address_width, sim_memory):
            self.data_width = data_width
            self.address_width = address_width
            self.sim_memory_size = len(sim_memory)
            self.fake_memory = Memory(width=data_width, depth=self.sim_memory_size, init=sim_memory)

            self.bus = ArbitraryWidthMemoryBus(data_width=data_width, address_width=address_width)

        def elaborate(self, platform):
            m = Module()

            write_ptr = Signal(range(self.sim_memory_size))

            m.submodules.read_port  = read_port  = self.fake_memory.read_port()
            bus = self.bus

            with m.If(bus.valid_in == 1):
                #m.d.comb += read_port.en.eq(0)
                m.d.comb += read_port.addr.eq(bus.r_addr)
                m.d.comb += bus.r_data.eq(read_port.data)
                m.d.sync += bus.valid_out.eq(1)


            return m



class DummyPlug(Elaboratable):

    #def __init__(self):



    def elaborate(self, platform):
        m = Module()

        m.submodules.FakeAWMem = FakeAWMem = ArbitraryWidthMemory(fake_data_width=7,
                            fake_address_width=8, initial_data=[0x23,0x45, 0x67, 0x89, 0xab,0xcd,0xef],
                            backing_memory_data_width=8, backing_memory_address_width=8)
        counter = Signal(8, reset=1)

#        with m.If(FakeAWMem.bus.ready_out == 1):
            #m.d.sync += counter.eq(counter+1)

        #with m.If(counter == 4):
        m.d.comb += FakeAWMem.bus.valid_in.eq(1)
        m.d.comb += FakeAWMem.bus.r_addr.eq(counter)


        return m



if __name__ == '__main__':
    baka =DummyPlug()
    main(baka)
    #platform.build(DummyPlug())
